import os
from dotenv import load_dotenv

load_dotenv()

MYSQL_DATABASE = os.environ["MYSQL_DATABASE"]
assert MYSQL_DATABASE

MYSQL_USER = os.environ["MYSQL_USER"]
assert MYSQL_USER

MYSQL_PASSWORD = os.environ["MYSQL_PASSWORD"]
assert MYSQL_PASSWORD

ALLOW_SIMULATION_NODE = os.environ["SIMULATION"] == "1"

AOI_OPERATOR_INITIAL_PASSWORD = os.environ["AOI_OPERATOR_INITIAL_PASSWORD"]
assert AOI_OPERATOR_INITIAL_PASSWORD

AOI_ADMINISTRATOR_INITIAL_PASSWORD = os.environ["AOI_ADMINISTRATOR_INITIAL_PASSWORD"]
assert AOI_ADMINISTRATOR_INITIAL_PASSWORD

STORAGE_SECRET = os.environ["STORAGE_SECRET"]
assert STORAGE_SECRET and len(STORAGE_SECRET) > 10



import bcrypt

from src.exceptions import AuthException


class Mixin:
    id: int
    hash: str

    @staticmethod
    def _hash_password(password: str) -> str:
        """Convert password to hash"""
        salt = bcrypt.gensalt()
        return bcrypt.hashpw(password.encode(), salt).decode()

    def set_password(self, password: str) -> None:
        """Override stored hash with new one"""
        self.hash = self._hash_password(password)

    def test_credentials(self, password: str) -> None:
        """Test password against stored hash"""

        print(password)
        print(self.hash)
        try:
            assert bcrypt.checkpw(password.encode(), self.hash.encode())
        except AssertionError as e:
            raise AuthException("Credential test failed") from e

    def assert_access(self, storage: dict):
        """Test storage record for access granted in past"""
        try:
            assert storage["access_allowed"]
            assert storage["accessor"] == self.id
        except (AssertionError, KeyError) as e:
            raise AuthException("Access assertion failed") from e

    def grant_access(self, storage: dict):
        """Push flags and metadata to app store to reflect access has been granted to accessor in past"""
        storage["access_allowed"] = True
        storage["accessor"] = self.id

    @staticmethod
    def revoke_access(storage: dict):
        """Remove access flags and metadata from session"""
        del storage["access_allowed"]
        del storage["accessor"]



from nicegui import ui, app

from src.enums import AccessorEnum
from src.models import Accessor
from src.exceptions import AuthException
from src.services.interface_web.views.utils import db_get_accessor


colors = dict(primary="#3A6B35", secondary="#CBD18F")

HOME_PAGE = "/"
ACCESS_PAGE = "/access"
INSPECTION_PROFILE_PAGE = "/inspection/profile"
INSPECTION_LIVE_PAGE = "/inspection/live"
DEVICES_PAGE = "/devices"
SETTINGS_PAGE = "/settings"


def inject_commons():
    ui.colors(**colors)


def _handle_logout_request():
    def logout():
        Accessor.revoke_access(app.storage.user)
        ui.open(ACCESS_PAGE)

    with ui.dialog() as dialog, ui.card():
        ui.label("You are about to logout. Are you sure?")
        with ui.row().classes("w-full justify-end"):
            ui.button("Cancel", on_click=dialog.close, color="white")
            ui.button("Confirm", on_click=logout, color="primary")

    dialog.open()


def inject_header():
    links_props = "flat square size=md"
    links_classes = "text-white"

    # TODO: adaptive
    with ui.header(fixed=True).classes("py-1"):
        ui.markdown("**AOI Portal** | Powered by ROS")
        ui.button("Overview", on_click=lambda: ui.open(HOME_PAGE)).classes(
            links_classes
        ).props(links_props)
        ui.button("Devices", on_click=lambda: ui.open(DEVICES_PAGE)).classes(
            links_classes
        ).props(links_props)
        ui.button("Settings", on_click=lambda: ui.open(SETTINGS_PAGE)).classes(
            links_classes
        ).props(links_props)
        ui.button(
            "Inspection profiles", on_click=lambda: ui.open(INSPECTION_PROFILE_PAGE)
        ).classes(links_classes).props(links_props)
        with ui.button(
            "Inspection", on_click=lambda: ui.open(INSPECTION_LIVE_PAGE)
        ).classes(links_classes).props(links_props):
            # ui.badge("offline", color="grey").classes('ml-1').props("rounded")
            ui.badge("online", color="red").classes("ml-1").props("rounded")
        ui.space()
        ui.button("Logout", on_click=_handle_logout_request).classes(
            links_classes
        ).props(links_props)


def inject_container():
    ui.left_drawer()
    ui.right_drawer()


def ensure_access_guard() -> None:
    operator = db_get_accessor(AccessorEnum.OPERATOR)
    if operator is None:
        raise AuthException("Access denied")

    administrator = db_get_accessor(AccessorEnum.ADMINISTRATOR)
    if administrator is None:
        raise AuthException("Access denied")

    try:
        operator.assert_access(app.storage.user)
    except AuthException:
        try:
            administrator.assert_access(app.storage.user)
        except AuthException:
            raise AuthException("Access denied")



from abc import ABC


class General(Exception, ABC):
    pass


class SecurityException(General, ABC):
    """Base exception for security issues"""

    pass


class AuthException(SecurityException):
    """Raised on auth attempt fail"""

    pass



import logging
from typing import Optional
from nicegui import ui, app
from fastapi.responses import RedirectResponse

from src.enums import AccessorEnum
from src.exceptions import AuthException
from src.services.interface_web.views.utils import db_get_accessor
from src.services.interface_web.views.common import inject_commons, HOME_PAGE

logger = logging.getLogger("ui.access")


def _handle_access_request(password_input: ui.input):
    operator = db_get_accessor(AccessorEnum.OPERATOR)
    if operator is None:
        return

    administrator = db_get_accessor(AccessorEnum.ADMINISTRATOR)
    if administrator is None:
        return

    try:
        operator.test_credentials(password=password_input.value)
    except AuthException as e:
        logger.info("Failed to test credentials against operator accessor")
        try:
            administrator.test_credentials(password=password_input.value)
        except AuthException as e:
            logger.info("Failed to test credentials against administrator accessor")
            ui.notify("Invalid credentials", type="negative")
            return
        else:
            # Allow admin access
            administrator.grant_access(app.storage.user)
            ui.open(HOME_PAGE)
            return
    else:
        # Allow operator access
        operator.grant_access(app.storage.user)
        ui.open(HOME_PAGE)
        return


def view() -> Optional[RedirectResponse]:
    operator = db_get_accessor(AccessorEnum.OPERATOR)
    if operator is None:
        return

    administrator = db_get_accessor(AccessorEnum.ADMINISTRATOR)
    if administrator is None:
        return

    # Do not allow logged in users to access page
    try:
        operator.assert_access(app.storage.user)
    except AuthException:
        pass
    else:
        return RedirectResponse(HOME_PAGE)

    try:
        administrator.assert_access(app.storage.user)
    except AuthException:
        pass
    else:
        return RedirectResponse(HOME_PAGE)

    # Render page
    ui.page_title("Access | AOI Portal")
    inject_commons()
    with ui.card().classes("absolute-center w-80"):
        with ui.row(wrap=False).classes("w-full justify-between items-center"):
            ui.markdown("**Enter credentials**")
            info = ui.button(icon="question_mark").props("flat round size=xs")
            info.tooltip("To access system please enter your credentials")
        password_input = ui.input(placeholder="Password", password=True).classes(
            "w-full"
        )
        ui.button(
            "Continue",
            color="primary",
            on_click=lambda: _handle_access_request(password_input),
        ).classes("w-full")

    with ui.header(fixed=True).classes("py-1"):
        ui.markdown("**AOI Portal** | Powered by ROS")
    with ui.footer(fixed=True).classes("py-2"):
        ui.label("Created by @Cherniaev.public")



from enum import Enum


class DefectTypeEnum(Enum):
    """Supported defect types (reflected in DB)"""

    MISSING_COMPONENT = 1
    WRONG_COMPONENT_ORIENTATION = 2
    TYPOGRAFY = 3


class RoleEnum(Enum):
    """Supported role types (reflected in DB)"""

    OPERATOR = 1
    ADMINISTRATOR = 2

class AccessorEnum(Enum):
    """Supported accessors (reflected in DB)"""

    OPERATOR = 1
    ADMINISTRATOR = 2



from datetime import datetime
from typing import Optional, List

from sqlalchemy import (
    String,
    ForeignKey,
    Numeric,
    DateTime,
    func,
    Boolean,
    Integer,
    MetaData,
)
from sqlalchemy import create_engine
from sqlalchemy.orm import Mapped, mapped_column, DeclarativeBase, relationship

from src.settings import (
    MYSQL_DATABASE,
    MYSQL_PASSWORD,
    MYSQL_USER,
)
from src.enums import DefectTypeEnum, RoleEnum, AccessorEnum
from src.mixins.control_zone import Mixin as ControlZoneMixin
from src.mixins.accessor import Mixin as AccessorMixin
from src.mixins.inspection_record import Mixin as InspectionRecordMixin
from src.mixins.template import Mixin as TemplateMixin
from src.mixins.camera import Mixin as CameraMixin
from src.mixins.inspection_profile import Mixin as InspectionProfileMixin


class Base(DeclarativeBase):
    pass


metadata_obj = MetaData()
engine = create_engine(
    f"mysql+pymysql://{MYSQL_USER}:{MYSQL_PASSWORD}@localhost/{MYSQL_DATABASE}"
)


class Role(Base):
    """Define accessor rights"""

    __tablename__ = "Role"
    metadata = metadata_obj

    id: Mapped[int] = mapped_column(
        primary_key=True, nullable=False, autoincrement=True
    )

    allow_system_view: Mapped[bool] = mapped_column(
        Boolean(), default=False, nullable=False
    )
    allow_inspection_log_view: Mapped[bool] = mapped_column(
        Boolean(), default=False, nullable=False
    )
    allow_inspection_details_view: Mapped[bool] = mapped_column(
        Boolean(), default=False, nullable=False
    )
    allow_inspection_flow_control: Mapped[bool] = mapped_column(
        Boolean(), default=False, nullable=False
    )
    allow_accessor_operations: Mapped[bool] = mapped_column(
        Boolean(), default=False, nullable=False
    )
    allow_device_operations: Mapped[bool] = mapped_column(
        Boolean(), default=False, nullable=False
    )
    allow_inspection_profile_operations: Mapped[bool] = mapped_column(
        Boolean(), default=False, nullable=False
    )
    allow_system_operations: Mapped[bool] = mapped_column(
        Boolean(), default=False, nullable=False
    )
    allow_statistics_view: Mapped[bool] = mapped_column(
        Boolean(), default=False, nullable=False
    )

    accessor_list: Mapped[List["Accessor"]] = relationship(back_populates="role")
    registry = RoleEnum


class Accessor(Base, AccessorMixin):
    __tablename__ = "Accessor"
    metadata = metadata_obj

    id: Mapped[int] = mapped_column(
        primary_key=True, nullable=False, autoincrement=True
    )

    title: Mapped[str] = mapped_column(String(50), nullable=False)
    description: Mapped[str] = mapped_column(String(200), nullable=False)

    role_id: Mapped[int] = mapped_column(ForeignKey("Role.id"), nullable=False)
    role: Mapped["Role"] = relationship()

    hash: Mapped[str] = mapped_column(String(60), nullable=False)
    registry = AccessorEnum


class DefectType(Base):
    """Define system wide known error types"""

    __tablename__ = "DefectType"
    metadata = metadata_obj

    id: Mapped[int] = mapped_column(primary_key=True, autoincrement=True)

    title: Mapped[str] = mapped_column(String(50), nullable=False)
    description: Mapped[str] = mapped_column(String(200), nullable=False)

    registry = DefectTypeEnum


class ControlTarget(Base):
    """
    Helper object to map defect type to search for in test image to the control zone.
    Multiple control targets are allowed for single control zone (unique)
    """

    __tablename__ = "ControlTarget"
    metadata = metadata_obj

    id: Mapped[int] = mapped_column(primary_key=True, autoincrement=True)

    defect_type_id: Mapped[int] = mapped_column(
        ForeignKey("DefectType.id"), nullable=False
    )
    defect_type: Mapped["DefectType"] = relationship()

    control_zone_id: Mapped[int] = mapped_column(
        ForeignKey("ControlZone.id"), nullable=False
    )
    control_zone: Mapped["ControlZone"] = relationship(
        back_populates="control_target_list"
    )


class ControlZone(Base, ControlZoneMixin):
    """
    Small zone on template where related defect type detection is conducted
    """

    __tablename__ = "ControlZone"
    metadata = metadata_obj

    id: Mapped[int] = mapped_column(primary_key=True, autoincrement=True)

    template_id: Mapped[int] = mapped_column(ForeignKey("Template.id"), nullable=False)
    template: Mapped["Template"] = relationship(back_populates="control_zone_list")

    top_left_x: Mapped[float] = mapped_column(Numeric(precision=10, scale=2))
    top_left_y: Mapped[float] = mapped_column(Numeric(precision=10, scale=2))

    bottom_right_x: Mapped[float] = mapped_column(Numeric(precision=10, scale=2))
    bottom_right_y: Mapped[float] = mapped_column(Numeric(precision=10, scale=2))

    control_target_list: Mapped[List["ControlTarget"]] = relationship(
        back_populates="control_zone", cascade="all, delete"
    )

    created_at: Mapped[datetime] = mapped_column(DateTime, server_default=func.now())
    created_by_accessor_id: Mapped[int] = mapped_column(
        ForeignKey("Accessor.id"), nullable=False
    )
    created_by: Mapped["Accessor"] = relationship()


class Defect(Base):
    """
    Helper to map defect type that was found to control zone. Multiple defects are allowed.
    """

    __tablename__ = "Defect"
    metadata = metadata_obj

    id: Mapped[int] = mapped_column(primary_key=True, autoincrement=True)

    defect_type_id: Mapped[int] = mapped_column(
        ForeignKey("DefectType.id"), nullable=False
    )
    defect_type: Mapped["DefectType"] = relationship()

    inspection_record_id: Mapped[int] = mapped_column(
        ForeignKey("InspectionRecord.id"), nullable=False
    )
    inspection_record: Mapped["InspectionRecord"] = relationship(
        back_populates="defect_list"
    )


class InspectionRecord(Base, InspectionRecordMixin):
    """
    Connect defect collection with template
    """

    __tablename__ = "InspectionRecord"
    metadata = metadata_obj

    id: Mapped[int] = mapped_column(primary_key=True, autoincrement=True)

    template_id: Mapped[int] = mapped_column(ForeignKey("Template.id"), nullable=False)
    template: Mapped["Template"] = relationship(back_populates="inspection_record_list")

    defect_list: Mapped[List["Defect"]] = relationship(
        back_populates="inspection_record", cascade="all, delete"
    )

    created_at: Mapped[datetime] = mapped_column(DateTime, server_default=func.now())


class Template(Base, TemplateMixin):
    """
    Main reference image. Aggregate control zones.
    """

    __tablename__ = "Template"
    metadata = metadata_obj

    id: Mapped[int] = mapped_column(primary_key=True, autoincrement=True)
    is_active: Mapped[bool] = mapped_column(Boolean(), default=False, nullable=False)

    image_uid: Mapped[str] = mapped_column(String(100), nullable=False)

    control_zone_list: Mapped[List["ControlZone"]] = relationship(
        back_populates="template", cascade="all, delete"
    )

    inspection_record_list: Mapped[List["InspectionRecord"]] = relationship(
        back_populates="template", cascade="all, delete"
    )

    inspection_profile_id: Mapped[int] = mapped_column(
        ForeignKey("InspectionProfile.id"), nullable=False
    )
    inspection_profile: Mapped["InspectionProfile"] = relationship(
        back_populates="template_list"
    )

    created_at: Mapped[datetime] = mapped_column(DateTime, server_default=func.now())
    created_by_accessor_id: Mapped[int] = mapped_column(
        ForeignKey("Accessor.id"), nullable=False
    )
    created_by: Mapped["Accessor"] = relationship()


class Camera(Base, CameraMixin):
    """
    Represent available cameras
    """

    __tablename__ = "Camera"
    metadata = metadata_obj

    id: Mapped[int] = mapped_column(primary_key=True, autoincrement=True)

    title: Mapped[str] = mapped_column(String(200), nullable=False)
    description: Mapped[str] = mapped_column(String(500), nullable=False)

    # TODO: insert validation ipv4
    ip_address: Mapped[str] = mapped_column(String(15), nullable=False)
    port: Mapped[str] = mapped_column(Integer, nullable=False)

    created_at: Mapped[datetime] = mapped_column(DateTime, server_default=func.now())
    created_by_accessor_id: Mapped[int] = mapped_column(
        ForeignKey("Accessor.id"), nullable=False
    )
    created_by: Mapped["Accessor"] = relationship()


class InspectionProfile(Base, InspectionProfileMixin):
    """
    Concrete instance of desired test configuration
    """

    __tablename__ = "InspectionProfile"
    metadata = metadata_obj

    id: Mapped[int] = mapped_column(primary_key=True, autoincrement=True)

    title: Mapped[str] = mapped_column(String(200), nullable=False)
    description: Mapped[str] = mapped_column(String(500), nullable=False)

    identification_code: Mapped[str] = mapped_column(String(100), nullable=False)

    # Point to all available templates
    template_list: Mapped[List["Template"]] = relationship(
        back_populates="inspection_profile", cascade="all, delete"
    )

    camera_id: Mapped[Optional[int]] = mapped_column(
        ForeignKey("Camera.id"), nullable=True
    )
    camera: Mapped[Optional["Camera"]] = relationship()

    created_at: Mapped[datetime] = mapped_column(DateTime, server_default=func.now())
    created_by_accessor_id: Mapped[int] = mapped_column(
        ForeignKey("Accessor.id"), nullable=False
    )
    created_by: Mapped["Accessor"] = relationship()



import numpy as np


class Mixin:
    def test(self) -> bool:
        raise NotImplemented()

    def start(self) -> bool:
        raise NotImplemented()

    def stop(self) -> bool:
        raise NotImplemented()

    def capture(self) -> np.ndarray:
        raise NotImplemented()



from nicegui import ui
from src.settings import STORAGE_SECRET

from src.services.interface_web.views.view_access import view as view_access
from src.services.interface_web.views.view_home import view as view_home

ui.page("/")(view_home)
ui.page("/access")(view_access)

ui.run(port=8000, storage_secret=STORAGE_SECRET)



"""Scripts create system wide known records in DB"""

from sqlalchemy.orm import Session

from src.models import *
from src.enums import *
from src.settings import (
    AOI_ADMINISTRATOR_INITIAL_PASSWORD,
    AOI_OPERATOR_INITIAL_PASSWORD,
)

if __name__ == "__main__":
    with Session(engine) as session:
        # Defect types
        dt_missing_component = DefectType(
            id=DefectTypeEnum.MISSING_COMPONENT.value,
            title="Missing component",
            description="Component is present on template, but is missing on tested image.",
        )

        dt_wrong_component_orientation = DefectType(
            id=DefectTypeEnum.WRONG_COMPONENT_ORIENTATION.value,
            title="Wrong component orientation",
            description="Component orientation is different against template.",
        )

        dt_typography = DefectType(
            id=DefectTypeEnum.TYPOGRAFY.value,
            title="Typografy",
            description="Typografy quality issues.",
        )

        # Roles
        r_operator = Role(
            id=RoleEnum.OPERATOR.value,
            allow_system_view=True,
            allow_inspection_log_view=True,
            allow_inspection_details_view=True,
            allow_inspection_flow_control=True,
            allow_accessor_operations=False,
            allow_device_operations=False,
            allow_inspection_profile_operations=False,
            allow_system_operations=False,
            allow_statistics_view=False,
        )

        r_administrator = Role(
            id=RoleEnum.ADMINISTRATOR.value,
            allow_system_view=True,
            allow_inspection_log_view=True,
            allow_inspection_details_view=True,
            allow_inspection_flow_control=True,
            allow_accessor_operations=True,
            allow_device_operations=True,
            allow_inspection_profile_operations=True,
            allow_system_operations=True,
            allow_statistics_view=True,
        )

        # Accessors
        a_operator = Accessor(
            id=AccessorEnum.OPERATOR.value,
            title="Operator",
            description="Operator is capable of basic sytem control including inspection requests.",
            role_id=RoleEnum.OPERATOR.value,
            hash=Accessor._hash_password(AOI_OPERATOR_INITIAL_PASSWORD),
        )
        a_administrator = Accessor(
            id=AccessorEnum.ADMINISTRATOR.value,
            title="Administrator",
            description="Administrator is granted full access to system including security section and inspection configuration.",
            role_id=RoleEnum.ADMINISTRATOR.value,
            hash=Accessor._hash_password(AOI_ADMINISTRATOR_INITIAL_PASSWORD),
        )

        session.add_all(
            [
                dt_missing_component,
                dt_wrong_component_orientation,
                dt_typography,
                r_operator,
                r_administrator,
                a_operator,
                a_administrator,
            ]
        )
        session.commit()



import logging
from typing import Optional

from nicegui import ui
from fastapi.responses import RedirectResponse

from src.exceptions import AuthException
from src.services.interface_web.views.common import (
    inject_commons,
    inject_header,
    inject_container,
    ACCESS_PAGE,
    ensure_access_guard,
)

logger = logging.getLogger("ui.home")


def view() -> Optional[RedirectResponse]:
    try:
        ensure_access_guard()
    except AuthException:
        return RedirectResponse(ACCESS_PAGE)

    # Render page
    ui.page_title("Home | AOI Portal")
    inject_commons()
    inject_header()
    inject_container()



import numpy as np


class Mixin:
    def set_top_left_coordinates(self, x: float, y: float) -> None:
        pass

    def set_bottom_right_coordinates(self, x: float, y: float) -> None:
        pass

    def inpaint(self, im: np.ndarray) -> np.ndarray:
        pass

    def crop(self, im: np.ndarray) -> np.ndarray:
        pass



import logging
from typing import Optional

from nicegui import ui
from sqlalchemy import select
from sqlalchemy.orm import Session

from src.enums import AccessorEnum
from src.models import engine, Accessor

logger = logging.getLogger("ui.utils")


def db_get_accessor(accessor_id: AccessorEnum) -> Optional[Accessor]:
    with Session(engine) as session:
        q = select(Accessor).where(Accessor.id == accessor_id.value)
        accessor = session.scalars(q).one_or_none()
        try:
            assert accessor is not None
        except AssertionError as e:
            logger.exception(e)
            ui.notify("Database inconsistency. Contact support.", type="warning")
            return
    return accessor



import numpy as np
from typing import List

class Mixin:
    image: np.ndarray

    def assign_image(self, im: np.ndarray) -> str:
        raise NotImplemented()

    def load_image(self):
        raise NotImplemented()

    def inpaint_zones(self, im: np.ndarray) -> np.ndarray:
        raise NotImplemented()

    def crop_zones(self, im: np.ndarray) -> List[np.ndarray]:
        raise NotImplemented()
